"
    Copyright (c) 2012-2016 Igor Stasenko
                            Martin McClure
                            Damien Pollet
                            Camillo Bruni
                            Guido Chari
                   2016-now Jan Vrany <jan.vrany [at] fit . cvut . cz>

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the 'Software'), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"
"{ Package: 'jv:dragonfly/asm' }"

"{ NameSpace: Smalltalk }"

AJOperand subclass:#AJMem
	instanceVariableNames:'size base index shift segmentPrefix hasLabel target displacement'
	classVariableNames:''
	poolDictionaries:'AJConstants'
	category:'AsmJit-Operands'
!

AJMem comment:'I am memory operand used in assembly instructions. I can be created from an immedate or a register.

Memory operands are used to read values indirectly from memory using certain offsets.

Example:
	asm := AJx86Assembler new.
	
	"create an memory operand on the address 1234"
	1234 asImm ptr
	
	"create a simple memory operand with RAX as base"
	asm RAX ptr.
	
	"the same with a 8 byte offset"
	asm RAX ptr + 8'
!

!AJMem class methodsFor:'documentation'!

copyright
"
    Copyright (c) 2012-2016 Igor Stasenko
                            Martin McClure
                            Damien Pollet
                            Camillo Bruni
                            Guido Chari
                   2016-now Jan Vrany <jan.vrany [at] fit . cvut . cz>

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the 'Software'), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"
! !

!AJMem class methodsFor:'instance creation'!

new
    "return an initialized instance"

    ^ self basicNew initialize.
! !

!AJMem methodsFor:'accessing'!

* aScale
    self scale: aScale
!

+ displacementOrIndex

    displacementOrIndex isInteger 
        ifTrue: [ 
            self displacement: (AJImmediate new ivalue: displacementOrIndex).
            ^ self ].
    
    displacementOrIndex isGeneralPurpose 
        ifTrue: [ 
            index := displacementOrIndex. 
            ^ self ].
    
     self error: 'Expected integer or general purpose register for memory displacement but got ', displacementOrIndex class name, '.'.
!

- aDisplacement

    aDisplacement isInteger 
        ifFalse: [ self error: 'Expected integer for memory displacement but got ', aDisplacement class name, '.' ].
    self displacement: (AJImmediate new ivalue: aDisplacement negated).
    ^ self
!

base
    "Answer the value of base"

    ^ base
!

base: anObject
    "Set the value of base"

    base := anObject
!

displacement
    "Answer the value of displacement"

    ^ displacement
!

displacement: anImm
    "Set the value of displacement"
    self assert: anImm isImm.
    displacement := anImm
!

hasLabel
    "Answer the value of hasLabel"
    ^ false
!

hasLabel: anObject
    "Set the value of hasLabel"

    hasLabel := anObject
!

index
    "Answer the value of index"

    ^ index
!

index: anIndex
    "Set the value of index, must be a general purpose register"
    self assert: (anIndex isGeneralPurpose).
    index := anIndex
!

scale: aScale
    "a valid scale values is 1 , 2 , 4 and 8"

    aScale = 1 ifTrue: [  shift := 0. ^ self ].	
    aScale = 2 ifTrue: [  shift := 1. ^ self ].	
    aScale = 4 ifTrue: [  shift := 2. ^ self ].	
    aScale = 8 ifTrue: [  shift := 3. ^ self ].	
        
    self error: 'invalid scale value'
!

segmentPrefix
    "Answer the value of segmentPrefix"

    ^ segmentPrefix
!

segmentPrefix: anObject
    "Set the value of segmentPrefix"

    segmentPrefix := anObject
!

shift
    "Answer the value of shift"

    ^ shift
!

shift: value
    "Set the value of shift"
    self assert: (value >=0 and: [ value < 4 ]).
    shift := value
!

size

    ^ size
!

size: anObject
    "Set the value of size"

    size := anObject
! !

!AJMem methodsFor:'emitting'!

emit32BitAbsoluteDisplacementModRM: emitter code: rCode
    self hasIndex
        ifTrue: [ 
            self assert: index index ~= RIDESP.	" ESP/RSP"
            emitter emitMod: 0 reg: rCode rm: 4.
            emitter emitScale: shift index: index index base: 5 ]
        ifFalse: [ emitter emitMod: 0 reg: rCode rm: 5 ].
        
    self hasLabel
        ifTrue: [ 
            "X86 uses absolute addressing model, all relative addresses will be
             relocated to absolute ones."
            "target is label"
            target
                addRelocationAt: emitter offset
                displacement: displacement
                absolute: true
                size: 4.
            emitter emitInt32: 0 ]
        ifFalse: [ 
            " Absolute address"
            displacement emitUsing: emitter size: 4 ]
!

emitBaseDisplacementModRM: emitter code: rCode
    | mod |
    
    self base isRip
        ifTrue: [ 
            emitter emitMod: 0 reg: rCode rm: 2r101.
            displacement emitUsing: emitter size: 4.
            ^ self ]. 
        
    mod := 0.
    displacement isZero
        ifFalse: [ 
            mod := displacement isInt8 ifTrue: [ 1 ] ifFalse: [ 2 ]].
        
    self base index == RIDESP
        ifTrue: [ 
            "ESP/RSP/R12"
            emitter emitMod: mod reg: rCode rm: RIDESP.
            emitter emitScale: 0 index: RIDESP base: RIDESP ]
        ifFalse: [ 
            (self base index ~= RIDEBP and: [ displacement isZero ])
                ifTrue: [ 
                    "just base, and not EBP/RBP/R13 "
                    ^ emitter emitMod: 0 reg: rCode rm: base index ].
            "force emitting displacement"
            mod = 0ifTrue: [ mod := 1 ].
            emitter emitMod: mod reg: rCode rm: base index ].
            
    mod = 1 ifTrue: [ displacement emitUsing: emitter size: 1 ].
    mod = 2 ifTrue: [ displacement emitUsing: emitter size: 4 ].
!

emitModRM: emitter code: rCode immSize: immSize
    "Receiver is memory location. rCode is a register number"

    "[base + displacement]"
    (self hasBase and: [ self hasIndex not ])
        ifTrue: [ ^ self emitBaseDisplacementModRM: emitter code: rCode ].	
    
    "[base + index * scale + displacement]"
    (self hasBase and: [ self hasIndex ])
        ifTrue: [ ^ self emitScaledBaseDisplacementModRM: emitter code: rCode ].
        
    " Address                       | 32-bit mode | 64-bit mode
   ------------------------------+-------------+---------------
   [displacement]                |   ABSOLUTE  | RELATIVE (RIP)
   [index * scale + displacemnt] |   ABSOLUTE  | ABSOLUTE (ZERO EXTENDED)
   In 32 bit mode is used absolute addressing model.
   In 64 bit mode is used relative addressing model together with absolute
   addressing one. The main problem is that if the instruction contains a SIB byte
   then relative addressing (RIP) is not possible. "
    emitter is32BitMode
        ifTrue: [ ^ self emit32BitAbsoluteDisplacementModRM: emitter code: rCode ].
            
    emitter is64BitMode
        ifTrue: [ self shouldBeImplemented ].
        
    self invalidInstruction
!

emitScaledBaseDisplacementModRM: emitter code: rCode

    self assert: index index ~= RIDESP.
    (base index ~= RIDEBP and: [ displacement isZero ])
        ifTrue: [ 
            emitter emitMod: 0 reg: rCode rm: 4.
            ^ emitter emitScale: shift index: index index base: base index ].
        
    displacement isInt8
        ifTrue: [ 
            emitter emitMod: 1 reg: rCode rm: 4.
            emitter emitScale: shift index: index index base: base index.
            displacement emitUsing: emitter size: 1 ]
        ifFalse: [ 
            emitter emitMod: 2 reg: rCode rm: 4.
            emitter emitScale: shift index: index index base: base index.
            displacement emitUsing: emitter size: 4 ].
    ^ self
! !

!AJMem methodsFor:'initialize-release'!

initialize
    displacement := AJImmediate new.
    shift := 0.
! !

!AJMem methodsFor:'printing'!

printOn: aStream
    self printAnnotationOn: aStream.
    aStream nextPutAll: 'mem['.
    base
        ifNotNil: [ 
            base printAsMemBaseOn: aStream.
            (index isNil and: [ displacement isNil ])
                ifFalse: [ aStream nextPutAll: ' + ' ] ].
    index
        ifNotNil: [ 
            aStream nextPutAll: index registerName.
            self printScaleOn: aStream.
            displacement ifNotNil: [ aStream nextPutAll: ' + ' ] ].
    displacement ifNotNil: [ aStream print: displacement ].
    aStream nextPut: $]
!

printScaleOn: aStream
    aStream nextPutAll: ' * '.
    (2 raisedToInteger: shift) printOn: aStream
! !

!AJMem methodsFor:'testing'!

hasBase
    ^ base notNil
!

hasIndex
    ^ index notNil
!

hasSegmentPrefix
    ^ segmentPrefix notNil
!

hasUpperBankIndex
    "True iff I have an index register, and it is one of r8-r15"

    ^ self hasIndex and: [ self index isUpperBank ]
!

isMem
    ^ true
!

isRip
    ^ self base isRip
!

isUpperBank
    "see `AJBaseReg >> #isUpperBank` "
    ^ base notNil and:[ base isUpperBank ]

    "Modified: / 16-06-2016 / 13:58:36 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

prohibitsRex 
    "Answer true if use of this operand requires that the instruction *not* have a REX prefix."

    ^ false
!

requiresRex
    "Answer true if use of this operand requires that the instruction have a REX prefix.
    For a memory reference, this is true if width of the transfer is 64, 
    or if either the base or index register is in the upper bank -- the
    use of a 64-bit base or index register is not enough by itself."

    ^ self is64 or: [ (self hasBase and: [ base isUpperBank ]) or: [ self hasIndex and: [ index isUpperBank ] ] ]
! !

