"
    Copyright (c) 2012-2016 Igor Stasenko
                            Martin McClure
                            Damien Pollet
                            Camillo Bruni
                            Guido Chari
                   2016-now Jan Vrany <jan.vrany [at] fit . cvut . cz>

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the 'Software'), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"
"{ Package: 'jv:dragonfly/asm' }"

"{ NameSpace: Smalltalk }"

Object subclass:#AJRoutineStackManager
	instanceVariableNames:'calls instructions assembler noStackFrame temps maxTemps
		extraStackBytes'
	classVariableNames:''
	poolDictionaries:''
	category:'AsmJit-StackManagement'
!

!AJRoutineStackManager class methodsFor:'documentation'!

copyright
"
    Copyright (c) 2012-2016 Igor Stasenko
                            Martin McClure
                            Damien Pollet
                            Camillo Bruni
                            Guido Chari
                   2016-now Jan Vrany <jan.vrany [at] fit . cvut . cz>

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the 'Software'), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"
! !

!AJRoutineStackManager class methodsFor:'instance creation'!

new
    "return an initialized instance"

    ^ self basicNew initialize.
! !

!AJRoutineStackManager methodsFor:'as yet unclassified'!

analyzeInstructions: anInstructions assembler: asm
    
    instructions := anInstructions.
    assembler := asm.
    
    instructions do: #prepareCallAlignments.
        
    calls do: [:callInfo |
        callInfo asm: assembler.
        callInfo emitAlignmentIfNeeded ].

    instructions do: [:each | each processTempsWith: self].

    self emitPrologue.

    ^ instructions 
!

emitEpilogue: popExtraBytes assembler: asm

    asm leave.
    
    popExtraBytes > 0 ifTrue: [
        asm ret: popExtraBytes asUImm
        ]
    ifFalse: [
        asm ret.
        ].
!

newCdeclCall 
    self stackFrameCheck.
    ^ calls add: (AJCdeclCallInfo new)
!

newStdCall
    self stackFrameCheck.
    ^ calls add: (AJStdCallCallInfo new)
!

noStackFrame
    noStackFrame := true.
!

releaseTemps: count
    temps := temps - count
!

reserveExtraBytesOnStack: numBytes
    
    self stackFrameCheck.
    
    extraStackBytes := numBytes 
!

reserveTemp: anAJReserveTemp
    self stackFrameCheck.
    temps := temps + 1.
    maxTemps := maxTemps max: temps .
    
    anAJReserveTemp operand: (assembler stackFrameValueAtOffset: extraStackBytes + (temps * assembler wordSize )).
!

stackFrameCheck

    noStackFrame ifTrue: [
        self error: 'Operation requires stack frame management to be enabled for generated code'
        ].
! !

!AJRoutineStackManager methodsFor:'emitting'!

emitPrologue

    noStackFrame == true ifTrue: [ ^self ]. 
    instructions do: [:each |
        each setPrologue: (assembler instructionsFor: [ | numBytes |
            assembler 
                push: assembler EBP;
                mov: assembler ESP to: assembler EBP.
                
            numBytes := extraStackBytes.
            numBytes := numBytes + (maxTemps * assembler wordSize ).
             
            numBytes > 0 ifTrue: [
                (assembler 
                    sub: assembler ESP with: numBytes) annotation:
                        extraStackBytes asString , ' extra bytes + ' ,
                        maxTemps asString , ' temps'
                ]			
            ]).
        ]
! !

!AJRoutineStackManager methodsFor:'initialize-release'!

initialize
    self reset.
    noStackFrame := false.
!

reset
    instructions := nil.
    assembler := nil.
    calls := OrderedCollection new.
    temps := maxTemps := extraStackBytes := 0.
! !

